"""
The pxpay payment processor involves call backs to the site which
result in an update on the status of an order during and after a
payment transaction.
"""
import logging
from Acquisition import aq_inner
from zope.component import getUtility

from Products.Five.browser import BrowserView
from Products.CMFCore.utils import getToolByName

from getpaid.core.interfaces import IOrderManager, IShoppingCartUtility, \
     workflow_states, IOrder

from getpaid.pxpay import parser
from getpaid.pxpay.interfaces import IPXPayStandardOptions, \
     IPXPayWebInterfaceGateway, IPXPayInvalidMessageError, \
     IPXPayNetworkError, ICheckoutContext
from getpaid.pxpay.exceptions import PXPayException, PXPayNetworkException, \
     PXPayInvalidMessageException
from getpaid.pxpay.config import RETURNED_TEST_CARD_NUMBER, TEST_SERVER_TYPE


log = logging.getLogger('getpaid.pxpay')

class ProcessResponse(BrowserView):
    """
    pxppay calls back on the url given in the initial request for
    payment. This url should be set to this view. This view is
    responsible for generating a ProcessResponse packet and sending
    this to pxpay to get confirmation details for the payment. see
    http://www.dps.co.nz/technical_resources/ecommerce_hosted/pxpay.html#ProcessResponse
    """

    def __init__(self, context, request):
        super(ProcessResponse, self).__init__(context, request)
        context = aq_inner(self.context)
        self.checkout_context = ICheckoutContext(context)
        self.store = self.checkout_context.store
        self.processor_options = IPXPayStandardOptions(self.store)
        self.pxpay_gateway = IPXPayWebInterfaceGateway(self.processor_options)

    def __call__(self):
        context = aq_inner(self.context)
        order_id = None
        if IOrder.providedBy(context):
            order_id = context.order_id
        encrypted_response = self.request.form.get('result', None)
        if not encrypted_response:
            raise PXPayException("There should be a result attribute in the form data for this view")
        process_response_message = parser.ReturnRequest()
        process_response_message.pxpay_user_id = self.processor_options.PxPayUserId
        process_response_message.pxpay_key = self.processor_options.PxPayKey
        process_response_message.response = encrypted_response

        state_valid, errors = process_response_message.state_validate()
        if not state_valid:
            raise PXPayException(errors)

        try:
            data = self.pxpay_gateway.send_message(process_response_message)
            log.info("About to send: %s" % process_response_message)
        except PXPayNetworkException:
            getUtility(IPXPayNetworkError)(self.context, self.request,
                                           order_id=order.order_id)
            return
        response_message = parser.ReturnResponse(data)
        log.info("Recieved: %s" % response_message)

        if not response_message.is_valid_response:
            getUtility(IPXPayInvalidMessageError)(self.context, self.request,
                                                  order_id=order_id,
                                                  message=response_message)
            return

        order_id = response_message.transaction_id
        order_manager = getUtility( IOrderManager )
        order = order_manager.get(order_id)

        if order is None:
            raise PXPayException("Order id " + order_id + " not found")

        # if you have Fail Proof Result Notification set on your pxpay
        # account then this callback will be called once or more by
        # pxpay and then also by your user when their browser is
        # redirected by pxpy
        current_state = order.finance_state
        f_states = workflow_states.order.finance

        if current_state == f_states.CHARGING:
            if self.determine_success(response_message):
                order.finance_workflow.fireTransition('charge-charging')
            else:
                order.finance_workflow.fireTransition('decline-charging')
        new_state = order.finance_state
        if new_state == f_states.CHARGED:
            self.destroy_cart()

        next_url = self.get_next_url(order)
        self.request.response.redirect(next_url)

    def determine_success(self, response_message):
        """
        Wrap up some logic to reconcile DPS transaction success in terms
        of whether we believe we are in Test mode or not
        """
        if response_message.transaction_successful:
            # Apparently all successful.
            # One last check, normally test card numbers with PXPay should only
            # work when you are using a development account.
            if response_message.transaction_card_number == RETURNED_TEST_CARD_NUMBER:
                # Someone has used the Test CC number...
                if self.processor_options.PxPayServerType != TEST_SERVER_TYPE:
                    # ...and we believe that we are in Production mode - log this:
                    log.info("Incorrect configuration? Test CC number was used for a successful transaction, but configuration indicates that we are not in Test mode: '%s'" % response_message)
            return True
        else:
            return False

    def destroy_cart(self):
        """
        time to destroy the cart
        """
        getUtility( IShoppingCartUtility ).destroy( self.context )


    def get_next_url(self, order):
        state = order.finance_state
        f_states = workflow_states.order.finance
        base_url = '/'.join((self.checkout_context.root_url,
                             '@@getpaid-order',
                             order.order_id))
        if not 'http://' in base_url:
            base_url = base_url.replace("https://", "http://")

        if state in (f_states.CANCELLED,
                     f_states.CANCELLED_BY_PROCESSOR,
                     f_states.PAYMENT_DECLINED):
            return base_url + '/@@getpaid-cancelled-declined'

        if state in (f_states.CHARGEABLE,
                     f_states.CHARGING,
                     f_states.REVIEWING,
                     f_states.CHARGED):
            return base_url + '/@@getpaid-thank-you?order_id=%s&finance_state=%s' %(order.order_id, state)



from getpaid.pxpay import _
from getpaid.pxpay.config import *
from zope.schema import vocabulary

def serverUrlsChoices( context ):
    return vocabulary.SimpleVocabulary.fromItems(
        [
         (_(u"Test Server"), TEST_SERVER_TYPE),
         (_(u"Production Server"), PRODUCTION_SERVER_TYPE),
        ]
       )

def currencyChoices( context ):
    return vocabulary.SimpleVocabulary.fromItems(
        [
         (_(u"Canadian Dollar"), u"CAD"),
         (_(u"Swiss Franc"), u"CHF"),
         (_(u"Euro"), u"EUR"),
         (_(u"French Franc"), u"FRF"),
         (_(u"United Kingdom Pound"), u"GBP"),
         (_(u"Hong Kong Dollar"), u"HKD"),
         (_(u"Japanese Yen"), u"JPY"),
         (_(u"New Zealand Dollar"), u"NZD"),
         (_(u"Singapore Dollar"), u"SGD"),
         (_(u"United States Dollar"), u"USD"),
         (_(u"Rand"), u"ZAR"),
         (_(u"Australian Dollar"), u"AUD"),
         (_(u"Samoan Tala"), u"WST"),
         (_(u"Vanuatu Vatu"), u"VUV"),
         (_(u"Tongan Pa'anga"), u"TOP"),
         (_(u"Solomon Islands Dollar"), u"SBD"),
         (_(u"Papua New Guinea Kina"), u"PNG"),
         (_(u"Malaysian Ringgit"), u"MYR"),
         (_(u"Kuwaiti Dinar"), u"KWD"),
         (_(u"Fiji Dollar"), u"FJD"),
        ]
       )



import decimal,operator
from exceptions import Exception

from zope import component
from zope import schema
from zope.formlib import form
from zc.table import column

from getpaid.core import interfaces, options
from getpaid.wizard import interfaces as wizard_interfaces
from Products.Five.browser.pagetemplatefile import ZopeTwoPageTemplateFile
from Products.PloneGetPaid.browser.checkout import CheckoutReviewAndPay, \
     sanitize_custom_widgets, null_condition
from Products.PloneGetPaid.browser import cart as cart_core
from Products.PloneGetPaid.interfaces import IGetPaidManagementOptions

from getpaid.pxpay.interfaces import IPXPayPaymentProcessor
from getpaid.pxpay.exceptions import PXPayException

from Products.PloneGetPaid.i18n import _

class PxPayCheckoutReviewAndPay( CheckoutReviewAndPay ):
    """
    Override the default PloneGetPaid CheckoutReviewAndPay step so
    that we can handle the details of the pxpay async processing of
    credit card details.
    """
    sections = ()

    template = ZopeTwoPageTemplateFile("templates/checkout-review-pay.pt")

    columns = [
        column.GetterColumn( title=_(u"Quantity"), getter=cart_core.LineItemColumn("quantity") ),
        column.GetterColumn( title=_(u"Name"), getter=cart_core.lineItemURL ),
        column.GetterColumn( title=_(u"Price"), getter=cart_core.lineItemPrice ),
        column.GetterColumn( title=_(u"Total"), getter=cart_core.lineItemTotal ),
       ]

    def customise_widgets(self,fields):
        pass

    def getSchemaAdapters( self ):
        return []

    def setUpWidgets( self, ignore_request=False ):
        self.adapters = self.adapters is not None and self.adapters or {}

        # grab all the adapters and fields from the entire wizard form
        # sequence (till the current step)
        adapters = self.wizard.data_manager.adapters
        fields   = self.wizard.data_manager.fields

        # display widgets for bill/ship address
        bill_ship_fields = []
        form_schemas = component.getUtility(
            interfaces.IFormSchemas)
        for i in (form_schemas.getInterface('billing_address'),
                  form_schemas.getInterface('shipping_address')):
            bill_ship_fields.append(fields.select(*schema.getFieldNamesInOrder(i)))
        # make copies of custom widgets.. (typically for edit, we want display)
        bill_ship_fields = sanitize_custom_widgets(
            reduce(operator.__add__,bill_ship_fields)
            )

        self.widgets = form.setUpEditWidgets(
            bill_ship_fields,  self.prefix, self.context, self.request,
            adapters=adapters, for_display=True, ignore_request=ignore_request
            )

    @form.action(_(u"Cancel"), name="cancel", validator=null_condition )
    def handle_cancel( self, action, data):
        url = self.context.portal_url.getPortalObject().absolute_url()
        url = url.replace("https://", "http://")
        return self.request.response.redirect(url)

    @form.action(_(u"Back"), name="back", validator=null_condition )
    def handle_back( self, action, data):
        self.next_step_name = wizard_interfaces.WIZARD_PREVIOUS_STEP

    @form.action(_(u"Make Payment"), name="make-payment" )
    def makePayment( self, action, data ):
        """ create an order, and submit to the pxpay processor """
        portal = self.context.portal_url.getPortalObject()
        manage_options = IGetPaidManagementOptions( portal )
        processor_name = manage_options.payment_processor
        if not processor_name:
            raise RuntimeError( "No Payment Processor Specified" )
        processor = component.getAdapter( portal,
                                          interfaces.IPaymentProcessor,
                                          processor_name )
        order = self.createOrder()
        order.processor_id = processor_name
        order.finance_workflow.fireTransition( "create" )
        order_manager = component.getUtility( interfaces.IOrderManager )
        order_manager.store( order )
        # the following will redirect to the pxpay web interface to
        # capture creditcard details
        if IPXPayPaymentProcessor.providedBy(processor):
            processor.authorize( order, None, self.request )
        else:
            raise PXPayException("This checkout implementation requires the PXPay payment processor.")




from zope.interface import Interface, Attribute
from zope import schema

from getpaid.core import interfaces

from getpaid.pxpay import _

class IPXPayStandardOptions(interfaces.IPaymentProcessorOptions):
    """
    PXPay Standard Options
    """
    PxPayServerType = schema.Choice(title = _(u"PXPay Website Payments Server"),
                                    vocabulary = "getpaid.pxpay.server_urls")

    PxPayUserId = schema.ASCIILine(
        title = _(u"PXPay UserId"),
        description = _(u"Enter your PXPay account UserId"),
        required = True,
       )

    PxPayKey = schema.ASCIILine(
        title = _(u"PXPay Key"),
        description = _(u"Enter the 64 character key that you were supplied for your PXPay account"),
        required = True,
      )

    MerchantReference = schema.ASCIILine(
        title = _(u"Merchant Reference"),
        description = _(u"Enter the 64 character merchant reference. This is free text to appear on transaction reports."),
        required = False,
      )

    PxPaySiteCurrency = schema.Choice(
        title = _(u"Site Currency"),
        vocabulary = "getpaid.pxpay.currencies",
       )


class IPXPayWebInterfaceGateway( Interface ):
    """
    A utility for connecting to and sending XML messages to the DPS
    pxpay web interface gateway.
    """

    def send_message(message, timeout):
        """
        Send a PXPay XML message to the pxpay web interface
        """

    def send_message_offline_testmode(message):
        """
        An offline mode that returns a canned test response to a
        send_message.
        """

class IPXPayInvalidMessageError( Interface ):
    """Utility to provide a hanlder for pxpay message errors,
    """

class IPXPayNetworkError( Interface ):
    """Utility to provide a handler for pxpay network errors.
    """


class IPXPayPaymentProcessor( interfaces.IPaymentProcessor ):
    """
    PXPay payment processor
    """

class ICheckoutContext(Interface):

    """
    The context where we want checkout to happen. Useful if you want
    to use a specific context with a particular portlet arrangement.
    """

    root_object = Attribute("the checkout root object")

    root_url = Attribute("the checkout root url")

    store = Attribute("the store object for this site")



import doctest
from zope.app.tests import placelesssetup
from zope.configuration.xmlconfig import XMLConfig
from Products.PloneGetPaid.config import PLONE3

from zope.component import getMultiAdapter, getUtility
from getpaid.core.interfaces import IOrderManager, IShoppingCartUtility, \
     IOrderWorkflowLog
from getpaid.core.order import Order
from getpaid.core.item import LineItem
from getpaid.pxpay.interfaces import IPXPayInvalidMessageError, \
     IPXPayNetworkError

# Standard options for DocTests
optionflags =  (doctest.ELLIPSIS |
                doctest.NORMALIZE_WHITESPACE)


def configurationSetUp(self):
    """Set up Zope 3 test environment
    """

    placelesssetup.setUp()

    # Ensure that the ZCML registrations in CMFonFive and PloneGetPaid are in effect
    # Also ensure the Five directives and permissions are available

    import Products.Five
    if not PLONE3:
        import Products.CMFonFive
    import Products.PloneGetPaid

    XMLConfig('configure.zcml', Products.Five)()
    XMLConfig('meta.zcml', Products.Five)()

    XMLConfig('configure.zcml', Products.PloneGetPaid)()

def configurationTearDown(self):
    """Tear down Zope 3 test environment
    """

    placelesssetup.tearDown()


class ContactInformation( object ):

    name = 'test user'
    email = 'test@xxx.xxx'


def create_test_order(context, order_id='1111111111'):
    """
    creates an order with a line item using a new cart.
    """
    item = LineItem()
    item.item_id = order_id
    item.name = "Jug of beer"
    item.cost = 5.00
    item.quantity = 5
    item.description = "Really nice beer"

    cart_util = getUtility(IShoppingCartUtility)
    cart = cart_util.get(context, create=True)
    cart[ item.item_id ] = item

    order = Order()
    order.order_id = order_id
    order.shopping_cart = cart
    order.processor_id = 'PXPay Processor'
    order.contact_information = ContactInformation()
    return order


def print_order_workflow_log(order):
    olog = IOrderWorkflowLog(order)
    for record in list(iter(olog)):
        print record.previous_state, ' --> ', record.new_state


def set_errors_to_raise_exceptions():
    invalidmessage_utility = getUtility(IPXPayInvalidMessageError)
    invalidmessage_utility.redirect = False
    networkerror_utility = getUtility(IPXPayNetworkError)
    networkerror_utility.redirect = False
